using System.Reflection;
using System.Runtime.CompilerServices;
using System.Runtime.InteropServices;

[assembly: AssemblyTitle ("Mono.Reflection.Tests")]
[assembly: AssemblyDescription ("")]
[assembly: AssemblyConfiguration ("")]
[assembly: AssemblyCompany ("Novell, Inc.")]
[assembly: AssemblyProduct ("Mono.Reflection.Tests")]
[assembly: AssemblyCopyright ("Copyright © Novell, Inc. 2009 - 2010")]
[assembly: AssemblyTrademark ("")]
[assembly: AssemblyCulture ("")]

[assembly: ComVisible (false)]

[assembly: Guid ("5f470471-07a4-4f0a-b33c-986b559528f6")]

[assembly: AssemblyVersion ("1.0.0.0")]
[assembly: AssemblyFileVersion ("1.0.0.0")]
