//
// DisassemblerTest.cs
//
// Author:
//   Jb Evain (jbevain@novell.com)
//
// (C) 2009 - 2010 Novell, Inc. (http://www.novell.com)
//
// Permission is hereby granted, free of charge, to any person obtaining
// a copy of this software and associated documentation files (the
// "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to
// permit persons to whom the Software is furnished to do so, subject to
// the following conditions:
//
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
// NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
// LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
// OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
// WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//

using System;
using System.IO;
using System.Reflection;
using System.Reflection.Emit;

using NUnit.Framework;

namespace Mono.Reflection {

	[TestFixture]
	public class BackingFieldResolverTest : BaseReflectionTest {

		[Test]
		public void StaticOptimizedProperty ()
		{
			AssertBackingField ("sopan", "SOPan");
		}

		[Test]
		public void StaticUnoptimizedProperty ()
		{
			AssertBackingField ("supan", "SUPan");
		}

		[Test]
		public void InstanceOptimizedProperty ()
		{
			AssertBackingField ("opan", "OPan");
		}

		[Test]
		public void InstanceUnoptimizedProperty ()
		{
			AssertBackingField ("upan", "UPan");
		}

		static void AssertBackingField (string field_name, string property_name)
		{
			var field = GetField (field_name);
			var property = GetProperty (property_name);

			Assert.AreEqual (field, property.GetBackingField ());
		}
	}
}
